#pragma once

#include <string>
#include <vector>
#include <array>
#include <memory>

#include <dune/grid/io/file/vtk/vtkwriter.hh>
#include <dune/grid/io/file/vtk/vtksequencewriter.hh>
#include <dune/geometry/genericgeometry/referenceelements.hh>

#include <dune/amdis/Initfile.hpp>
#include <dune/amdis/common/Loops.hpp>

namespace AMDiS
{

  template <class Traits>
  class FileWriter
  {
  private: // typedefs and static constants
    
    using Mesh     = typename Traits::Mesh;
    using MeshView = typename Mesh::LeafGridView;
    
    /// Dimension of the mesh
    static constexpr int dim = Traits::dim;
    
    /// Dimension of the world
    static constexpr int dow = Traits::dimworld;
    
    /// Number of problem components
    static constexpr int nComponents = Traits::nComponents;
      
      
  public:
    
    /// Constructor.
    FileWriter(std::string base, MeshView const& meshView, std::vector<std::string> const& names_)
      : meshView(meshView)
      , names(names_)
    {
      filename = "solution";
      Parameters::get(base + "->filename", filename);
      dir = "output";
      Parameters::get(base + "->output directory", dir);
      
      vtkWriter = make_shared<Dune::VTKWriter<MeshView>>(meshView);
      vtkSeqWriter = make_shared<Dune::VTKSequenceWriter<MeshView>>(vtkWriter, filename, dir, "");
    }

    
    /// default write method for time-depended data
    template <class SystemVectorType>
    void write(double time, SystemVectorType const& solutions)
    {
      vtkWriter->clear();
      // copy dofvector to vertex data
      for_<0, nComponents>([this, &solutions](const auto _i) 
      {
        this->dofVector2vertexVector(solutions[_i], std::get<_i>(data_vectors));
        vtkSeqWriter->addVertexData(std::get<_i>(data_vectors), names[_i]);
      });
      vtkSeqWriter->write(time/*, Dune::VTK::appendedraw*/);
    }
    

    /// default write method for stationary data
    template <class SystemVectorType>
    void write(SystemVectorType const& solutions)
    {
      vtkWriter->clear();
      // copy dofvector to vertex data
      for_<0, nComponents>([this, &solutions](const auto _i) 
      {
        this->dofVector2vertexVector(solutions[_i], std::get<_i>(data_vectors));
        vtkWriter->addVertexData(std::get<_i>(data_vectors), names[_i]);
      });
      vtkWriter->pwrite(filename, dir, "" /*, Dune::VTK::appendedraw*/);
    }
    
      
  protected:
    
    template <class DOFVector, class Vector>
    void dofVector2vertexVector(DOFVector const& dofvector, Vector& data)
    {
      using Geometry = typename MeshView::template Codim<0>::Geometry;
      using RefElements = Dune::ReferenceElements<typename Geometry::ctype, Geometry::mydimension>;
      
      data.resize(meshView.size(dim));
      
      auto const& indexSet = meshView.indexSet();
      auto const& feSpace = dofvector.getFeSpace();
      
      auto localView = feSpace.localView();
      auto localIndexSet = feSpace.localIndexSet();
      
      // copy data to P1-vector
      for (auto const& element : elements(meshView)) {
        localView.bind(element);
        localIndexSet.bind(localView);
        auto const& localBasis = localView.tree().finiteElement().localBasis();
        
        auto const& refElement = RefElements::general(element.type());
        
        std::vector<Dune::FieldVector<double,1> > shapeValues;
        
        size_t nVertices = element.subEntities(dim);
        for (size_t i = 0; i < nVertices; ++i) {
          auto const& v = element.template subEntity<dim>(i);
          auto pos = refElement.position(i, dim);
          
          localBasis.evaluateFunction(pos, shapeValues);
          
          size_t idx = indexSet.index(v);
          data[idx] = 0.0;
          for (size_t j = 0; j < shapeValues.size(); ++j) {
            const auto global_idx = localIndexSet.index(j);
            data[idx] += dofvector[global_idx] * shapeValues[j];
          }
        }
      }
    }

  private:
    MeshView const& meshView;
    shared_ptr<Dune::VTKWriter<MeshView>> vtkWriter;
    shared_ptr<Dune::VTKSequenceWriter<MeshView>> vtkSeqWriter;
    
    std::vector<std::string> names;
    std::array<std::vector<double>, nComponents> data_vectors;
    
    std::string filename;
    std::string dir;
  };

} // end namespace AMDiS
