#pragma once

#include <string>
#include <stdio.h>  /* defines FILENAME_MAX */
#include <cctype>

#ifdef WINDOWS
    #include <direct.h>
    #define getCurrentDir _getcwd
#else
    #include <unistd.h>
    #define getCurrentDir getcwd
#endif
 
#include <dune/amdis/common/String.hpp> 

namespace AMDiS
{
  // A minimalistic filesystem class
  class Path
  {
  public:
    Path(std::string p)
    {
      split(p, pathParts);
      path_ = normalize(pathParts);
    }
    
    // return filename part of path
    std::string filename() const
    {
      return pathParts.back();
    }
    
    // get filename extension
    std::string extension() const
    {
      auto fn = filename();
      auto pos = fn.find_last_of('.');
      return fn.substr(pos+1);
    }
    
    // get current working directory
    std::string currentPath() const
    {
      char cwd_[FILENAME_MAX];
      getCurrentDir(cwd_, sizeof(cwd_));
      std::string cwd(cwd_);
      return trim(cwd);
    }
    
    std::string const& path() const
    {
      return path_;
    }
    
    // return normalized version of absolute path
    std::string absolute() const
    {
      std::string fullPath = "";
      if (isAbsolute(path_))
        fullPath = path_;
      else
        fullPath = currentPath() + "/" + path_;
      
      std::vector<std::string> parts;
      split(fullPath, parts);
      return normalize(parts);
    }
  
    bool isAbsolute(std::string p) const
    {
      if (p[0] == '/')
        return true;
      
      // c:\ or z:/
      if (std::isalpha(p[0]) && p[1] == ':' && (p[2] == '/' || p[2] == '\\'))
        return true;
      
      return false;
    }
    
  protected:
    std::string normalize(std::vector<std::string> const& parts) const
    {
      auto it = parts.begin();
      std::string p = *it;
      for (++it; it != parts.end(); ++it)
        p += sep + normalize(*it);
      return p;
    }
    
#ifdef WINDOWS
    std::string normalize(std::string const& part) const
    {
      return to_lower(part);
    }
#else
    std::string const& normalize(std::string const& part) const
    {
      return part;
    }
#endif  
    
    static void split(std::string const& p, std::vector<std::string>& parts)
    {
      using Tokenizer = boost::tokenizer<boost::char_separator<char>>;

      boost::char_separator<char> sep("/\\");
      Tokenizer tokens(p + "/" + p, sep);
      bool relative = true;
      for (auto token : tokens) {
        trim(token);
        if (!token.empty() && token != "." && token != "..") {
          parts.push_back(token);
          relative = false;
        } else if (token == ".." && relative) {
          parts.pop_back();
        }
      }
    }
    
  private:
    std::string path_;
    std::vector<std::string> pathParts;
    
#ifdef WINDOWS
    static constexpr char sep = '\\';
#else
    static constexpr char sep = '/';
#endif    
  };

} // end namespace AMDiS
