#pragma once

#include <string>

#include <dune/amdis/Initfile.hpp>

namespace AMDiS
{
  /// Class that stores information about the solution process, like tolerances 
  /// and iteration counts and is passed to the solver and filled there.
  class SolverInfo
  {
  public:
    /// The constructor reads needed parameters and sets solver \p prefix.
    /**
     * Reads parameters for a solver with name 'NAME':
     *   NAME->absolute tolerance   \ref aTol
     *   NAME->relative tolerance   \ref rTol
     *   NAME->info                 \ref info
     *   NAME->break if tolerance not reached \ref breakTolNotReached
    **/
    explicit SolverInfo(std::string prefix)
    {
      Parameters::get(prefix + "->abolute tolerance", aTol);
      Parameters::get(prefix + "->relative tolerance", rTol);
      Parameters::get(prefix + "->info", info);
      Parameters::get(prefix + "->break if tolerance not reached", breakTolNotReached);
    }

    /** \name getting methods
     * \{
     */

    /// Returns \ref aTol
    double getAbsTolerance() const
    {
      return aTol;
    }

    /// Returns \ref rTol
    double getRelTolerance() const
    {
      return rTol;
    }

    /// Returns error code in last run of an iterative solver
    int getErrorCode() const
    {
      return error;
    }

    /// Returns info
    int getInfo() const
    {
      return info;
    }

    /// Returns \ref absResidual
    double getAbsResidual() const
    {
      return absResidual;
    }

    /// Returns \ref relResidual
    double getRelResidual() const
    {
      return relResidual;
    }
    
    /// Returns the initfile \ref prefix
    std::string getPrefix() const
    {
      return prefix;
    }

    /// Returns \ref createMatrixData
    bool doCreateMatrixData() const
    {
      return createMatrixData;
    }

    /// Returns \ref storeMatrixData
    bool doStoreMatrixData() const
    {
      return storeMatrixData;
    }
    
    bool doBreak() const
    {
      return breakTolNotReached && (
          (aTol > 1.e-30 && absResidual > aTol) || 
          (rTol > 1.e-30 && relResidual > rTol) );
    }

    /** \} */


    /** \name setting methods
     * \{
     */

    /// Sets \ref aTol
    void setAbsTolerance(double tol)
    {
      aTol = tol;
    }

    /// Sets \ref rTol
    void setRelTolerance(double tol)
    {
      rTol = tol;
    }

    /// Sets \ref aTol
    void setAbsResidual(double r)
    {
      absResidual = r;
    }

    /// Sets \ref rTol
    void setRelResidual(double r)
    {
      relResidual = r;
    }

    /// Sets \ref info
    void setInfo(int i)
    {
      info = i;
    }

    /// Sets \ref error
    void setError(int e)
    {
      error = e;
    }

    /// Sets \ref createMatrixData
    void setCreateMatrixData(bool b)
    {
      createMatrixData = b;
    }

    /// Sets \ref storeMatrixData
    void setStoreMatrixData(bool b)
    {
      storeMatrixData = b;
    }
    
    

    /** \} */
    
  private:
    /// The initfile prefix to read parameters
    std::string prefix;
    
    /// The abolute tolerance
    double aTol = 0;
    
    /// The relative tolerance
    double rTol = 1.e-6;
    
    /// Throw an error if tolerance could not be reached
    bool breakTolNotReached = false;
    
    /// The solver verbosity level
    int info = 0;
    
    /// The absolute residual, default (-1): not set
    double absResidual = -1.0;
    
    /// The relative residual, relative to the two_norm(b), default (-1): not set
    double relResidual = -1.0;
    
    /// The error-code, default (-1): not set
    int error = -1;
    
    /// If true, the matrix will be initialized and the 
    /// corresponding runner of the system receives the 
    /// matrix in the init() method.
    bool createMatrixData = true;
    
    /// If false, the exit() method of the runner will be called.
    bool storeMatrixData = false;
  };

} // end namespace AMDiS
