#pragma once

#include <dune/amdis/linear_algebra/istl/ISTL_Solver.hpp>
#include <dune/amdis/linear_algebra/istl/Preconditioner.hpp>

namespace AMDiS
{
  template <class Matrix, class VectorX, class VectorB>
  class ISTLRunner
    : public RunnerInterface<Matrix, VectorX, VectorB>
  {
    using LinOperator = Dune::MatrixAdapter<Matrix, VectorX, VectorB>;
    using BaseSolver  = Dune::InverseOperator<VectorX, VectorB>;
    
    using PreconRunner = ISTLPreconditioner<Matrix, VectorX, VectorB>;
    using PreconBase   = PreconditionerInterface<Matrix, VectorX, VectorB>;

  public:
    ISTLRunner(std::string solverName, std::string prefix)
      : solverName(solverName)
      , prefix(prefix)
      , solverCreator(prefix)
    {
      // get creator string for preconditioner
      std::string preconName = "no";
      Parameters::get(prefix + "->left precon", preconName);
      if (preconName.empty() || preconName == "no")
        Parameters::get(prefix + "->right precon", preconName);
      if (preconName.empty() || preconName == "no")
        Parameters::get(prefix + "->precon->name", preconName);
      
      precon = make_shared<PreconRunner>(preconName, prefix + "->precon");
    }
    
    
    /// Implementes \ref RunnerInterface::init()
    virtual void init(Matrix const& A) override
    {
      precon->init(A);
      
      Matrix& _A = const_cast<Matrix&>(A); // Unschoen!!!
      linOperator = make_shared<LinOperator>(_A);
      solver = solverCreator.create(solverName, *linOperator, *precon);
    }

    /// Implementes \ref RunnerInterface::exit()
    virtual void exit() override 
    {
      solver.reset();
      linOperator.reset();
      
      precon->exit();
      precon.reset();
    }

    /// Implementes \ref RunnerInterface::solve()
    virtual int solve(Matrix const& A, VectorX& x, VectorB const& b, 
                      SolverInfo& solverInfo) override 
    {
      // storing some statistics
      Dune::InverseOperatorResult statistics;

      // solve the linear system
      VectorB _b = b;
      solver->apply(x, _b, statistics);
      
      solverInfo.setRelResidual(statistics.reduction);
    }
    
    /// Implementes \ref RunnerInterface::getLeftPrecon()
    virtual shared_ptr<PreconBase> getLeftPrecon() override
    {
      return preconAdapter;
    }
                      
  private:
    std::string solverName;
    std::string prefix;
    
    ISTLSolverCreator<Matrix, VectorX, VectorB> solverCreator;
    
    shared_ptr<PreconRunner> precon;
    shared_ptr<LinOperator>  linOperator;
    shared_ptr<BaseSolver>   solver;
  };

} // end namespace AMDiS
