#pragma once

#include <dune/amdis/CreatorMap.hpp>
#include <dune/amdis/linear_algebra/mtl/ITL_Solver.hpp>

namespace AMDiS
{
  /// Adds default creators for linear solvers based on `BlockMTLMatrix`.
  /**
   * Adds creators for block matrix aware solvers.
   * - *cg*: conjugate gradient method, \see cg_solver_type
   * - *cgs*: stabilized conjugate gradient mtheod, \see cgs_solver_type
   * - *bcgs*: stabilized bi-conjugate gradient method, \see bicgstab_type
   * - *tfqmr*: Transposed-Free Quasi-Minimal Residual method, \see tfqmr_solver_type
   * - *bcgsl*: stabilized BiCG(ell) method, \see bicgstab_ell_type
   * - *gmres*: Generalized minimal residula method, \see gmres_type
   * - *fgmres*: Flexible GMRES, \see fgmres_type
   * - *minres*: Minimal residul method, \see minres_solver_type
   * - *gcr*: Generalized conjugate residual method, \see gcr_type
   * - *umfpack*: external UMFPACK solver, \see UmfpackRunner
   **/
  template <class SubMatrix, size_t _N, size_t _M, class Vector>
  class DefaultCreators<LinearSolverInterface<BlockMTLMatrix<SubMatrix, _N, _M>, Vector, Vector>>
  {
    using Matrix = BlockMTLMatrix<SubMatrix, _N, _M>;
    using SolverBase = LinearSolverInterface<Matrix, Vector, Vector>;
    
    template <class ITLSolver>
    using SolverCreator
      = typename LinearSolver<Matrix, Vector, 
          KrylovRunner<ITLSolver, Matrix, BaseVector_t<Vector>>>::Creator;
    
    using UmfpackSolverCreator
      = typename LinearSolver<Matrix, Vector, 
          UmfpackRunner<Matrix, BaseVector_t<Vector>>>::Creator;
          
    using Map = CreatorMap<SolverBase>;
    
  public:
    static void init()
    { 
      auto cg = new SolverCreator<cg_solver_type>;
      Map::addCreator("cg", cg);
      
      auto cgs = new SolverCreator<cgs_solver_type>;
      Map::addCreator("cgs", cgs);
        
      auto bcgs = new SolverCreator<bicgstab_type>;
      Map::addCreator("bicgstab", bcgs);
      Map::addCreator("bcgs", bcgs);
        
      auto tfqmr = new SolverCreator<tfqmr_solver_type>;
      Map::addCreator("tfqmr", tfqmr);
        
      auto bcgsl = new SolverCreator<bicgstab_ell_type>;
      Map::addCreator("bicgstab_ell", bcgsl);
      Map::addCreator("bcgsl", bcgsl);
        
      auto gmres = new SolverCreator<gmres_type>;
      Map::addCreator("gmres", gmres);
        
      auto fgmres = new SolverCreator<fgmres_type>;
      Map::addCreator("fgmres", fgmres);
        
      auto minres = new SolverCreator<minres_solver_type>;
      Map::addCreator("minres", minres);
        
      auto gcr = new SolverCreator<gcr_type>;
      Map::addCreator("gcr", gcr);
    
      auto umfpack = new UmfpackSolverCreator;
      Map::addCreator("umfpack", umfpack);
      Map::addCreator("direct", umfpack);
      
      Map::addCreator("default", gmres);
    }
  };

} // end namespace AMDiS
