#pragma once

#include <array>
#include <type_traits>

#include <boost/numeric/mtl/utility/irange.hpp>

#include <dune/amdis/common/Mpl.hpp>
#include <dune/amdis/linear_algebra/LinearAlgebraBase.hpp>
#include <dune/amdis/linear_algebra/mtl/MTLDenseVector.hpp>

namespace AMDiS
{
  /// A simple block-vector class
  template <class MTLVector, size_t _N>
  class BlockMTLVector
    : public std::array<MTLVector, _N>
  {
  public:
    /// The type of the blocks
    using BaseVector = MTLVector;
    
    /// The index/size - type
    using size_type  = typename MTLVector::size_type;
    
    /// The type of the elements of the MTLVector
    using value_type = typename MTLVector::value_type;
  };
  
  
  namespace Impl
  {
    /// Specialization of Impl::BaseVector from \file LinearAlgebraBase.hpp
    template <class MTLVector, size_t N>
    struct BaseVector<BlockMTLVector<MTLVector, N>>
    {
      using type = MTLVector;
    };
  }
  

  /// Return the number of overall rows of a BlockMTLVector
  template <class MTLVector, size_t _N>
  inline size_t num_rows(BlockMTLVector<MTLVector, _N> const& vec)
  {
    size_t nRows = 0;
    for_<0, _N>([&](const auto _i) {
      nRows += num_rows(std::get<_i>(vec));
    });
    return nRows;
  }

  /// Return the number of overall columns of a BlockMTLVector
  template <class MTLVector, size_t _N>
  inline size_t num_cols(BlockMTLVector<MTLVector, _N> const& vec)
  {
    return 1;
  }

  /// Return the size, i.e. rows*columns of a BlockMTLVector
  template <class MTLVector, size_t _N>
  inline size_t size(BlockMTLVector<MTLVector, _N> const& vec)
  {
    return num_rows(vec);
  }

  /// Nullify a BlockMTLVector, i.e. nullify each block.
  template <class MTLVector, size_t _N>
  inline void set_to_zero(BlockMTLVector<MTLVector, _N>& vec) 
  {
    for_<0, _N>([&](const auto _i) {
      set_to_zero(std::get<_i>(vec));
    });
  }
  
  
  /// A wrapper, that creates a contiguos vector corresponding to a block-vector
  /// and copy the value on construction and eventually back on destruction, if 
  /// required.
  template <class BlockVector, class Vector, class NonConstBlockVector>
  class BlockVectorWrapper;
  
  // specialization for BlockMTLVector
  template <class BlockVector, class Vector, class MTLVector, size_t _N>
  class BlockVectorWrapper<BlockVector, Vector, BlockMTLVector<MTLVector, _N>>
  {    
    static_assert( std::is_same< std::remove_const_t<BlockVector>, BlockMTLVector<MTLVector, _N> >::value, 
                   "This specialization is for BlockMTLVectors only.");
    
  public:
    BlockVectorWrapper(BlockVector& blockVector, 
                       bool copyBack = !std::is_const<BlockVector>::value)
      : blockVector(blockVector)
      , vector(num_rows(blockVector))
      , copyBack(copyBack)
    {
      assignTo();
    }
    
    ~BlockVectorWrapper()
    {
      if (copyBack)
        assignFrom(bool_<!std::is_const<BlockVector>::value>());
    }
    
    /// Return a reference to the block-vector
    BlockVector const&  getBlockVector() const  { return blockVector; }
    
    /// Return a reference to the contiguose-vector
    Vector&       getVector()       { return vector; }
    Vector const& getVector() const { return vector; }

  private:
    /// copy from block-vector to vector
    void assignTo()
    {
      size_t start = 0;
      for (size_t r = 0; r < _N; ++r) {
        size_t finish = start + num_rows(blockVector[r]);
        mtl::irange range(start, finish);
        
        vector[range] = blockVector[r];
        start = finish;
      }
    }
    
    /// do not copy vector to block-vector since this is const
    template <bool Copy> 
    std::enable_if_t<!Copy> assignFrom(bool_<Copy>) { /* Do nothing */ }
    
    /// copy from vector to block-vector
    template <bool Copy> 
    std::enable_if_t<Copy> assignFrom(bool_<Copy>)
    {      
      size_t start = 0;
      for (size_t r = 0; r < _N; ++r) {
        size_t finish = start + num_rows(blockVector[r]);
        mtl::irange range(start, finish);
        
        blockVector[r] = vector[range];
        start = finish;
      }
    }
    
  private:
    BlockVector& blockVector;
    Vector vector;
    
    /// Copy data back to block-vector on destruction
    bool copyBack;
  };
  
  
  // Specialization for non-block-vectors
  template <class BlockVector, class Vector>
  class BlockVectorWrapper<BlockVector, Vector, Vector>
  {
    static_assert( std::is_same< std::remove_const_t<BlockVector>, Vector >::value, 
                   "This specialization is for contiguose vectors only.");
  public:
    BlockVectorWrapper(BlockVector& vector, bool = false)
      : vector(vector)
    {}
    
    /// Return a reference to the vector
    BlockVector const& getBlockVector() const { return vector; }
    BlockVector&       getVector()            { return vector; }
    BlockVector const& getVector()      const { return vector; }
    
  private:
    BlockVector& vector;
  };
  
  template <class BlockVector, class Vector>
  using BlockVectorWrapper_t
    = BlockVectorWrapper<BlockVector, Vector, std::remove_const_t<BlockVector>>;

  
  template <class BlockVector, class Vector = MTLDenseVector<typename BlockVector::value_type>>
  BlockVectorWrapper_t<BlockVector, Vector>
  blockWrapper(BlockVector& bvec)
  {
    return {bvec};
  }
    

} // end namespace AMDiS
