#pragma once

#include <string>
#include <memory>

#include <boost/numeric/mtl/matrix/compressed2D.hpp>
#include <boost/numeric/mtl/matrix/inserter.hpp>
#include <boost/numeric/mtl/utility/property_map.hpp>
#include <boost/numeric/mtl/utility/range_wrapper.hpp>

#include <dune/amdis/Log.hpp>
#include <dune/amdis/common/ClonablePtr.hpp>

namespace AMDiS
{
  /// \brief The basic container that stores a base matrix and a corresponding 
  /// row/column feSpace.
  template <class RowFeSpaceType, class ColFeSpaceType, 
            class ValueType = double>
  class DOFMatrix
  {
  public:
    /// The type of the finite element space / basis of the row
    using RowFeSpace = RowFeSpaceType;
    
    /// The type of the finite element space / basis of the column
    using ColFeSpace = ColFeSpaceType;
    
    /// The matrix type of the underlying base matrix
    using BaseMatrix = mtl::compressed2D<ValueType>;
    
    /// The index/size - type
    using size_type  = typename BaseMatrix::size_type;
    
    /// The type of the elements of the DOFMatrix
    using value_type = ValueType;
    
    /// The underlying field-type (typically the same as \ref value_type)
    using field_type = typename BaseMatrix::value_type;
    
    /// The type of the inserter used when filling the matrix. NOTE: need not be public
    using Inserter = mtl::mat::inserter<BaseMatrix, mtl::operations::update_plus<value_type>>;
    
  public:
    /// Constructor. Constructs new BaseMatrix.
    DOFMatrix(RowFeSpace const& rowFeSpace, 
              ColFeSpace const& colFeSpace, 
              std::string name)
      : rowFeSpace(rowFeSpace)
      , colFeSpace(colFeSpace)
      , name(name)
      , matrix(ClonablePtr<BaseMatrix>::make())
    {}
    
    /// Constructor. Takes pointer of data-matrix.
    DOFMatrix(RowFeSpace const& rowFeSpace, 
              ColFeSpace const& colFeSpace, 
              std::string name, 
              BaseMatrix& matrix_ref)
      : rowFeSpace(rowFeSpace)
      , colFeSpace(colFeSpace)
      , name(name)
      , matrix(matrix_ref)
    {}
    
    /// Return the row-basis \ref rowFeSpace of the matrix
    RowFeSpace const& getRowFeSpace() const
    {
      return rowFeSpace;
    }
    
    /// Return the col-basis \ref colFeSpace of the matrix
    ColFeSpace const& getColFeSpace() const
    {
      return colFeSpace;
    }
    
    /// Return a reference to the data-matrix \ref matrix
    BaseMatrix& getMatrix()
    {
      return *matrix;
    }
    
    /// Return a reference to the data-matrix \ref matrix
    BaseMatrix const& getMatrix() const
    {
      return *matrix;
    }
    
    /// Return the size of the row \ref feSpace
    size_type N() const
    {
      return rowFeSpace.size();
    }
    
    /// Return the size of the column \ref feSpace
    size_type M() const
    {
      return colFeSpace.size();
    }
    
    /// Return the \ref name of this matrix
    std::string getName() const
    {
      return name;
    }
    
    /// \brief Returns an update-proxy of the inserter, to inserte/update a value at 
    /// position (\p r, \p c) in the matrix. Need an initialized inserter, that can
    /// be created by \ref init and must be closed by \ref finish after insertion.
    auto operator()(size_type r, size_type c)
    {
      AMDIS_TEST_EXIT_DBG( initialized, "Inserter not initialized!");
      AMDIS_TEST_EXIT_DBG( r < N() && c < M() , 
          "Indices out of range [0," << N() << ")x[0," << M() << ")" );
      return (*inserter)[r][c];
    }
    
    /// Create inserter. Assumes that no inserter is currently active on this matrix.
    void init(bool prepareForInsertion)
    {
      AMDIS_TEST_EXIT(!initialized && !inserter, "Matrix already in insertion mode!");
      
      calculateSlotSize();
      matrix->change_dim(rowFeSpace.size(), colFeSpace.size());
      if (prepareForInsertion) {
        set_to_zero(*matrix);
        inserter = new Inserter(*matrix, slot_size);
        initialized = true;
      }
    }

    /// Delete inserter -> finish insertion. Must be called in order to fill the 
    /// final construction of the matrix.
    void finish()
    {
      delete inserter;
      inserter = NULL;
      initialized = false;
    }
    
    /// Return the number of nonzeros in the matrix
    size_t getNnz() const
    {
      return matrix->nnz();
    }
    
    /// \brief Deletes all rows with \p dirichletNodes != 0 and if \p apply is true, adds
    /// a one on the diagonal of the matrix.
    void clearDirichletRows(std::vector<char> const& dirichletNodes, bool apply)
    {    
      // Define the property maps
      auto row   = mtl::mat::row_map(*matrix); 
      auto col   = mtl::mat::col_map(*matrix);
      auto value = mtl::mat::value_map(*matrix); 

      // iterate over the matrix    
      for (auto r : mtl::major_of(*matrix)) {   // rows or columns
        for (auto i : mtl::nz_of(r)) {          // non-zeros within
          if (!dirichletNodes[row(i)])
            break;
          
          // set identity row
          value(i, (apply && row(i) == col(i) ? value_type(1) : value_type(0)) );
        }
      }
    }
    
  protected:
    // Estimates the slot-size used in the inserter to reserve enough space per row.
    void calculateSlotSize()
    {
      slot_size = 0;

      if (num_rows(*matrix) != 0)
        slot_size = int(double(matrix->nnz()) / num_rows(*matrix) * 1.2);
      if (slot_size < MIN_NNZ_PER_ROW)
        slot_size = MIN_NNZ_PER_ROW;
    }
      
  private:
    /// The minimal number of nonzeros per row
    static constexpr int MIN_NNZ_PER_ROW = 20;
    
    /// The finite element space / basis of the row
    RowFeSpace const& rowFeSpace;
    
    /// The finite element space / basis of the column
    ColFeSpace const& colFeSpace;
    
    /// The name of the DOFMatrix
    std::string name;
    
    /// The data-matrix (can hold a new BaseMatrix or a pointer to external data
    ClonablePtr<BaseMatrix> matrix;
    
    /// A pointer to the inserter. Created in \ref init and destroyed in \ref finish
    Inserter* inserter = NULL;
    
    /// A flag that indicates whether the matrix is in insertion mode
    bool initialized = false;
    
    /// The size of the slots used to insert values per row
    int slot_size = 20;
    
    // friend class declarations
    template <class, class> friend class SystemMatrix;
  };

} // end namespace AMDiS
