/** \file ITL_Runner.h */

#pragma once

// MTL4 headers
#include <boost/numeric/itl/itl.hpp>
#include <boost/numeric/mtl/mtl.hpp>

// AMDiS headers
#include <dune/amdis/linear_algebra/PreconditionerInterface.hpp>
#include <dune/amdis/linear_algebra/RunnerInterface.hpp>
#include <dune/amdis/linear_algebra/SolverInfo.hpp>

#include <dune/amdis/linear_algebra/mtl/ITL_Preconditioner.hpp>
#include <dune/amdis/linear_algebra/mtl/BITL_Preconditioner.hpp>

namespace AMDiS
{
  /** \ingroup Solver
   *
   * \brief
   * Wrapper class for different MTL4 itl-solvers. These solvers
   * are parametrized by Matrix- and Vector.
   **/
  template <class ITLSolver, class Matrix, class Vector>
  class KrylovRunner : public RunnerInterface<Matrix, Vector>
  {
    using Super = RunnerInterface<Matrix, Vector>;
    using PreconBase = typename Super::PreconBase;
    
  public:
    /// Constructor.
    KrylovRunner(std::string prefix)
      : itlSolver(prefix)
    {
      initPrecon(prefix);
      
      Parameters::get(prefix + "->max iteration", maxIter);
      Parameters::get(prefix + "->print cycle", printCycle);
    }    
    
    /// Implements \ref RunnerInterface::getLeftPrecon(), Returns \ref L
    virtual shared_ptr<PreconBase> getLeftPrecon() override
    {
      return L;
    }

    /// Implements \ref RunnerInterface::getRightPrecon(), Returns \ref R
    virtual shared_ptr<PreconBase> getRightPrecon() override
    {
      return R;
    }
    
    /// Set a new left preconditioner \ref L
    void setLeftPrecon(shared_ptr<PreconBase> precon)
    {
      L = precon;
    }
    
    /// Set a new right preconditioner \ref R
    void setRightPrecon(shared_ptr<PreconBase> precon)
    {
      R = precon;
    }

    /// Implementation of \ref RunnerInterface::init()
    virtual void init(Matrix const& A) override
    {
      L->init(A);
      R->init(A);
    }
    
    /// Implementation of \ref RunnerInterface::exit()
    virtual void exit() override
    {
      L->exit();
      R->exit();
    }

    /// Implementation of \ref RunnerInterface::solve()
    virtual int solve(Matrix const& A, Vector& x, Vector const& b, 
                      SolverInfo& solverInfo) override
    {
      AMDIS_TEST_EXIT(L, "There is no left preconditioner");
      AMDIS_TEST_EXIT(R, "There is no right preconditioner");

      auto r = Vector(b);
      if (two_norm(x) != 0)
        r -= A * x;
      
      // print information about the solution process
      itl::cyclic_iteration<double> iter(r, maxIter, solverInfo.getRelTolerance(), 
                                                     solverInfo.getAbsTolerance(),
                                                     printCycle);
      iter.set_quite(solverInfo.getInfo() == 0);
      iter.suppress_resume(solverInfo.getInfo() == 0);

      int error = itlSolver(A, x, b, *L, *R, iter);
      solverInfo.setAbsResidual(iter.resid());
      solverInfo.setRelResidual(iter.relresid());

      return error;
    }

  private:
    /// Create left/right preconditioners from parameters given in the init-file
    void initPrecon(std::string prefix)
    {      
      // Creator for the left preconditioner
      std::string leftPrecon = "", rightPrecon = "";
      Parameters::get(prefix + "->left precon", leftPrecon);
      Parameters::get(prefix + "->right precon", rightPrecon);
      
      auto leftCreator
        = CreatorMap<PreconBase>::getCreator(leftPrecon, prefix + "->left precon");
      auto rightCreator
        = CreatorMap<PreconBase>::getCreator(rightPrecon, prefix + "->right precon");
        
      L = leftCreator->create();
      R = rightCreator->create();
    }

  private:
    /// The itl solver object that performes the actual solve
    ITLSolver itlSolver;
    
    /// Pointer to the left and right preconditioner
    shared_ptr<PreconBase> L, R;
    
    /// The maximal number of iterations
    size_t maxIter = 1000;
    
    /// Print solver residuals every \ref printCycle 's iteration
    size_t printCycle = 100;
    
  };

} // end namespace AMDiS
