/** \file LinearSolver.h */

#pragma once

// AMDiS includes
#include <dune/amdis/CreatorInterface.hpp>
#include <dune/amdis/Log.hpp>
#include <dune/amdis/common/Timer.hpp>
#include <dune/amdis/linear_algebra/LinearSolverInterface.hpp>
#include <dune/amdis/linear_algebra/mtl/BlockMTLVector.hpp>

namespace AMDiS
{
  /** \ingroup Solver
   *
   * \brief
   * Wrapper class for various MTL4 solvers. These solvers
   * are parametrized by MatrixType and VectorType. The different
   * algorithms, like krylov subspace methods or other external
   * solvers where MTL4 provides an interface, can be assigned
   * by different Runner objects.
   **/
  template <class Matrix, class Vector, class Runner>
  class LinearSolver 
    : public LinearSolverInterface<Matrix, Vector, Vector>
  {
    using Self = LinearSolver;
    using Super = LinearSolverInterface<Matrix, Vector, Vector>;
    using RunnerBase = typename Super::RunnerBase;
    
  public:
    /// A creator to be used instead of the constructor.
    struct Creator : CreatorInterfaceName<Super>
    {
      virtual shared_ptr<Super> create(std::string prefix) override
      { 
        return make_shared<Self>(prefix); 
      }
    };
    
  public:
    /// Constructor
    LinearSolver(std::string prefix)
      : runner(make_shared<Runner>(prefix))
    {}
    
    /// Implements \ref LinearSolverInterface::getRunner()
    virtual shared_ptr<RunnerBase> getRunner() override
    {
      return runner;
    }

  private:
    /// Implements \ref LinearSolverInterface::solveSystemImpl()
    virtual void solveImpl(Matrix const& A, Vector& x, Vector const& b,
                           SolverInfo& solverInfo) override
    {
      Timer t;
      if (solverInfo.doCreateMatrixData()) {
        // init matrix or wrap block-matrix or ...
        runner->init(A);
      }
      
      // wrappers copy the data back on destruction
      auto X = blockWrapper(x);
      auto B = blockWrapper(b);
      
      if (solverInfo.getInfo() > 0) {
        AMDIS_MSG("fill MTL4 matrix needed " << t.elapsed() << " seconds");
      }
      
      int error = runner->solve(A, X.getVector(), B.getVector(), solverInfo);
      solverInfo.setError(error);
      
      if (!solverInfo.doStoreMatrixData())
        runner->exit();
    }

  private:
    /// redirect the implementation to a runner. Implements a \ref RunnerInterface
    shared_ptr<Runner> runner;
  };

} // end namespace AMDiS
