#pragma once

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <vector>
#include <type_traits>

#include <dune/istl/bvector.hh>

#include <dune/amdis/OperatorTermBase.hpp>
#include <dune/amdis/common/Traits.hpp>

namespace AMDiS 
{
  /** 
    * \addtogroup OperatorTerms
    * @{
    **/

  namespace Impl 
  {
    template <class F, int dow>
    using CoordsFunctorResult
      = std::result_of<F(Dune::FieldVector<double, dow>)>;
      
    template <class F, int dow>
    constexpr bool CallableDow =
      Concepts::Callable<F, Dune::FieldVector<double, dow>>;
      
#ifndef AMDIS_DOW
    template <class F>
    using CoordsFunctorResult_t = typename std::conditional_t<
      CallableDow<F, 1>, CoordsFunctorResult<F, 1>, std::conditional_t<
      CallableDow<F, 2>, CoordsFunctorResult<F, 2>, std::conditional_t<
      CallableDow<F, 3>, CoordsFunctorResult<F, 3>, Id<void>> > >::type;
#else
    template <class F>
    using CoordsFunctorResult_t = typename std::conditional_t<
      CallableDow<F, AMDIS_DOW>, CoordsFunctorResult<F, AMDIS_DOW>, Id<void>>::type;
#endif
  }
  
  
  /// \brief An expression that evaluates to the current coordinate of a dof or 
  /// quadrature point with given index. \see eval.
  template <class Functor>
  class CoordsTerm
  {
  public:
    using value_type = Impl::CoordsFunctorResult_t<Functor>;
    static_assert( !std::is_same<value_type, void>::value, 
                   "Template `Functor` can not be used as a functor.");
    
    template <class F,
      class = std::enable_if_t<Concepts::Compatible<Functor, F>> >
    CoordsTerm(F&& f, int degree = 1)
      : fct(std::forward<F>(f))
      , degree(degree)
    {}
    
    /// \brief Evaluates the functor \ref fct at all quadrature points and 
    /// stores the result in a vector \ref values.
    template <class Element, class PointList>
    void init(Element const& element, 
              PointList const& points)
    {
      AMDIS_FUNCNAME("CoordsTerm::init()");
      values.resize(points.size());
      for (size_t iq = 0; iq < points.size(); ++iq)
        values[iq] = fct(element.geometry().global(points[iq].position()));
    }
    
    value_type const& operator[](size_t iq) const
    {
      return values[iq];
    }
    
    int getDegree() const
    {
      return degree;
    }
      
  private:
    Functor fct;
    int degree;
    
    std::vector<value_type> values;
  };
  
  
  /// Generator function for \ref CoordsTerm expressions
  template <class F>
  CoordsTerm< std::decay_t<F> > eval(F&& f) { return {std::forward<F>(f)}; }
  
  /** @} **/
  
} // end namespace AMDiS
