#pragma once

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <algorithm>
#include <vector>
#include <type_traits>

#include <dune/istl/bvector.hh>

#include <dune/amdis/OperatorTermBase.hpp>
#include <dune/amdis/common/Traits.hpp>
#include <dune/amdis/utility/GetDegree.hpp>

namespace AMDiS 
{
  /** 
    * \addtogroup OperatorTerms
    * @{
    **/

  
  /// \brief An expression that evalues a DOFVector at a given element, either on the
  /// dofs or on the quadrature points. \see valueOf.
  template <class DOFVectorType>
  class DOFVectorTerm
  {      
  public:
    using value_type = Value_t<DOFVectorType>;
    
    DOFVectorTerm(DOFVectorType const& dofvector, double factor = 1.0)
      : vector(dofvector.getVector())
      , factor(factor)
      , localView(dofvector.getFeSpace().localView())
      , localIndexSet(dofvector.getFeSpace().localIndexSet())
    {}
    
    template <class Element, class PointList>
    void init(Element const& element, 
              PointList const& points)
    {
      AMDIS_FUNCNAME("DOFVectorTerm::init()");
      localView.bind(element);
      localIndexSet.bind(localView);
      
      const auto& localFiniteElem = localView.tree().finiteElement();
      const size_t nBasisFct = localFiniteElem.size();
      
      std::vector<Dune::FieldVector<double,1> > shapeValues(nBasisFct);
      std::vector<value_type> localVec(nBasisFct);
      
      for (size_t j = 0; j < nBasisFct; ++j) {
        const auto global_idx = localIndexSet.index(j);
        localVec[j] = factor * vector[global_idx];
      }
      
      values.resize(points.size());
      for (size_t iq = 0; iq < points.size(); ++iq) {
        localFiniteElem.localBasis().evaluateFunction(points[iq].position(), shapeValues);
        value_type result = 0;
        for (size_t j = 0; j < shapeValues.size(); ++j)
          result += localVec[j] * shapeValues[j];
        values[iq] = result;
      }
    }
    
    value_type const& operator[](size_t iq) const
    {
      return values[iq];
    }
    
    int getDegree() const
    {
      return degree;
    }
      
  private:
    typename DOFVectorType::BaseVector const& vector;
    double factor;
    
    using Basis = typename DOFVectorType::FeSpace;
    typename Basis::LocalView localView;
    typename Basis::LocalIndexSet localIndexSet;
    
    int degree = getPolynomialDegree<Basis>;
    
    std::vector<value_type> values;
  };

  
  /// Generator function for \ref DOFVectorTerm expressions.
  /** Generates a \ref DOFVectorTerm, that takes a \ref DOFVector \p vector, 
   *  an optional factor \p factor the value of the DOFVector at the quadrature
   *  points are scaled with.
   **/
  template <class DOFVectorType>
  DOFVectorTerm<std::decay_t<DOFVectorType>>
  valueOf(DOFVectorType&& vector, double factor = 1.0)
  {
    return {std::forward<DOFVectorType>(vector), factor};
  }
  
  
// -----------------------------------------------------------------------------
  
  
  /// \brief An expression that evalues a DOFVector at a given element, either on the
  /// dofs or on the quadrature points, and applies a functor to the value. 
  /// \see valueOfFunc.
  template <class DOFVectorType, class Func>
  class DOFVectorFuncTerm
  {      
  public:
    using value_type = Value_t<DOFVectorType>;
    
    template <class F_>
    DOFVectorFuncTerm(DOFVectorType const& dofvector, F_&& f, int f_deg)
      : vector(dofvector.getVector())
      , f(std::forward<F_>(f))
      , localView(dofvector.getFeSpace().localView())
      , localIndexSet(dofvector.getFeSpace().localIndexSet())
      , f_deg(f_deg)
    {}
    
    template <class Element, class PointList>
    void init(Element const& element, 
              PointList const& points)
    {
      AMDIS_FUNCNAME("DOFVectorFuncTerm::init()");
      localView.bind(element);
      localIndexSet.bind(localView);
      
      const auto& localFiniteElem = localView.tree().finiteElement();
      const size_t nBasisFct = localFiniteElem.size();
      
      std::vector<Dune::FieldVector<double,1> > shapeValues(nBasisFct);
      std::vector<value_type> localVec(nBasisFct);
      
      for (size_t j = 0; j < nBasisFct; ++j) {
        const auto global_idx = localIndexSet.index(j);
        localVec[j] = vector[global_idx];
      }
      
      values.resize(points.size());
      for (size_t iq = 0; iq < points.size(); ++iq) {      
        localFiniteElem.localBasis().evaluateFunction(points[iq].position(), shapeValues);
        value_type result = 0;
        for (size_t j = 0; j < shapeValues.size(); ++j)
          result += localVec[j] * shapeValues[j];
        
        values[iq] = f(result);
      }
    }
    
    value_type const& operator[](size_t iq) const
    {
      return values[iq];
    }
    
    int getDegree() const
    {
      return degree * f_deg;
    }
      
  private:
    typename DOFVectorType::BaseVector const& vector;
    Func f;
    
    using Basis = typename DOFVectorType::FeSpace;
    typename Basis::LocalView localView;
    typename Basis::LocalIndexSet localIndexSet;
    
    int degree = getPolynomialDegree<Basis>;
    int f_deg;
    
    std::vector<value_type> values;
  };

    
  /// Generator function for \ref DOFVectorFuncTerm expressions.
  /** Generates a \ref DOFVectorFuncTerm, that takes a \ref DOFVector \p vector, 
   *  a functor \p f that operators on the value type of the DOFVector, and 
   *  optionally a degree argument \p deg that represents an (approximate)
   *  polynomial degree of the functor, for choosing the right quadrature degree.
   **/
  template <class DOFVectorType, class F>
  DOFVectorFuncTerm<std::decay_t<DOFVectorType>, std::decay_t<F>>
  valueOfFunc(DOFVectorType&& vector, F&& f, int deg = 1)
  {
    return {std::forward<DOFVectorType>(vector), std::forward<F>(f), deg};
  }
  
  
// -----------------------------------------------------------------------------
  
  
  /// \brief An expression that evalues the gradient of a DOFVector at a given 
  /// element, either on the DOFs or on the quadrature points. \see gradientOf.
  template <class DOFVectorType>
  class GradientTerm
  {      
    using Basis = typename DOFVectorType::FeSpace::LocalView::Tree::FiniteElement::Traits::LocalBasisType;
    using Jacobian = typename Basis::Traits::JacobianType;
    
  public:    
    using value_type = Jacobian;
    using field_type = Value_t<DOFVectorType>;
    
    GradientTerm(DOFVectorType const& dofvector, double factor = 1.0)
      : vector(dofvector.getVector())
      , factor(factor)
      , localView(dofvector.getFeSpace().localView())
      , localIndexSet(dofvector.getFeSpace().localIndexSet())
    {}
    
    template <class Element, class PointList>
    void init(Element const& element, 
              PointList const& points)
    {
      AMDIS_FUNCNAME("DOFVectorTerm::init()");
      localView.bind(element);
      localIndexSet.bind(localView);
      
      const auto& localFiniteElem = localView.tree().finiteElement();
      const size_t nBasisFct = localFiniteElem.size();
      
      std::vector<Jacobian> shapeGradients(nBasisFct);
      std::vector<field_type> localVec(nBasisFct);
      
      for (size_t j = 0; j < nBasisFct; ++j) {
        const auto global_idx = localIndexSet.index(j);
        localVec[j] = factor * vector[global_idx];
      }
      
      values.resize(points.size());
      for (size_t iq = 0; iq < points.size(); ++iq) {
        localFiniteElem.localBasis().evaluateJacobian(points[iq].position(), shapeGradients);
        value_type result = 0;
        for (size_t j = 0; j < shapeGradients.size(); ++j)
          result += localVec[j] * shapeGradients[j];
        values[iq] = result;
      }
    }
    
    value_type const& operator[](size_t iq) const
    {
      return values[iq];
    }
    
    int getDegree() const
    {
      return Math::max(0, degree-1);
    }
      
  private:
    typename DOFVectorType::BaseVector const& vector;
    double factor;
    
    using FeSpace = typename DOFVectorType::FeSpace;
    typename FeSpace::LocalView localView;
    typename FeSpace::LocalIndexSet localIndexSet;
    
    int degree = getPolynomialDegree<FeSpace>;
    
    std::vector<value_type> values;
  };

  
  /// Generator function for an expression of gradients of \ref DOFVectors
  /** Generates a \ref GradientTerm, that takes a \ref DOFVector \p vector, 
   *  and an optional factor \p factor the value of the gradient of the
   *  DOFVector at the quadrature points are scaled with.
   **/
  template <class DOFVectorType>
  GradientTerm<std::decay_t<DOFVectorType>>
  gradientOf(DOFVectorType&& vector, double factor = 1.0)
  {
    return {std::forward<DOFVectorType>(vector), factor};
  }
  
  
  /// \brief An expression that evalues a partial derivative of a  DOFVector 
  /// at a given element, either on the DOFs or on the quadrature points. \see derivativeOf.
  template <class DOFVectorType>
  class DerivativeTerm
  {      
    using Basis = typename DOFVectorType::FeSpace::LocalView::Tree::FiniteElement::Traits::LocalBasisType;
    using Jacobian = typename Basis::Traits::JacobianType;
    using Derivative = typename Basis::Traits::RangeType;
    
  public:
    using value_type = Derivative;
    using field_type = Value_t<DOFVectorType>;
    
    DerivativeTerm(DOFVectorType const& dofvector, int component, double factor = 1.0)
      : vector(dofvector.getVector())
      , component(component)
      , factor(factor)
      , localView(dofvector.getFeSpace().localView())
      , localIndexSet(dofvector.getFeSpace().localIndexSet())
    {
      direction[0] = component;
    }
    
    template <class Element, class PointList>
    void init(Element const& element, 
              PointList const& points)
    {
      AMDIS_FUNCNAME("DOFVectorTerm::init()");
      localView.bind(element);
      localIndexSet.bind(localView);
      
      const auto& localFiniteElem = localView.tree().finiteElement();
      const size_t nBasisFct = localFiniteElem.size();
      
//       std::vector<Derivative> shapeDerivative(nBasisFct);
      std::vector<Jacobian> shapeJacobian(nBasisFct);
      std::vector<field_type> localVec(nBasisFct);
      
      for (size_t j = 0; j < nBasisFct; ++j) {
        const auto global_idx = localIndexSet.index(j);
        localVec[j] = factor * vector[global_idx];
      }
      
      values.resize(points.size());
      for (size_t iq = 0; iq < points.size(); ++iq) {
//         localFiniteElem.localBasis().template evaluate<1>(direction, points[iq].position(), shapeDerivative);
        localFiniteElem.localBasis().evaluateJacobian(points[iq].position(), shapeJacobian);
        value_type result = 0;
        for (size_t j = 0; j < nBasisFct; ++j)
          result += localVec[j] * shapeJacobian[j][0][component];
        values[iq] = result;
      }
    }
    
    value_type const& operator[](size_t iq) const
    {
      return values[iq];
    }
    
    int getDegree() const
    {
      return Math::max(0, degree-1);
    }
      
  private:
    typename DOFVectorType::BaseVector const& vector;
    int component;
    double factor;
    
    std::array<int, 1> direction;
    
    using FeSpace = typename DOFVectorType::FeSpace;
    typename FeSpace::LocalView localView;
    typename FeSpace::LocalIndexSet localIndexSet;
    
    int degree = getPolynomialDegree<FeSpace>;
    
    std::vector<value_type> values;
  };

  
  /// Generator function for an expression of partial derivatives of \ref DOFVectors
  /** Generates a \ref DerivativeTerm, that takes a \ref DOFVector \p vector, 
   *  a index representing the component of the gradient that should be returned,
   *  and an optional factor \p factor the value of the partial derivative of the
   *  DOFVector at the quadrature points are scaled with.
   **/
  template <class DOFVectorType>
  DerivativeTerm<std::decay_t<DOFVectorType>>
  derivativeOf(DOFVectorType&& vector, int component, double factor = 1.0)
  {
    return {std::forward<DOFVectorType>(vector), component, factor};
  }
  
  /** @} **/
  
} // end namespace AMDiS
