#pragma once

#include <nanoflann.hpp>

#include <cstdlib>
#include <iostream>


namespace AMDiS { namespace extensions {

using namespace nanoflann;

// =====================================================================================
typedef WorldVector<double> PointType;
typedef boost::tuple<const MacroElement*, int, unsigned long> DataType;
typedef std::vector<PointType> VectorOfPointsType;
typedef std::vector<DataType> VectorOfDataType;
// =====================================================================================


  /** A simple vector-of-vectors adaptor for nanoflann, without duplicating the storage.
    *  The i'th vector represents a point in the state space.
    *
    *  \tparam dim If set to >0, it specifies a compile-time fixed dimensionality for the points in the data set, allowing more compiler optimizations.
    *  \tparam value_type The type of the point coordinates (typically, double or float).
    *  \tparam Distance The distance metric to use: nanoflann::metric_L1, nanoflann::metric_L2, nanoflann::metric_L2_Simple, etc.
    *  \tparam index_type The type for indices in the KD-tree index (typically, size_t of int)
    */
  template < class FeSpace >
  struct KDTreeMeshAdaptor
  {
    using MeshView = typename FeSpace::GridView;
    using SizeType = typename FeSpace::LocalIndexSet::size_type;
    using ValueType = double;
    
    static constexpr dimension = MeshView::dimension;
    static constexpr dimensionworld = MeshView::dimensionworld;
    
    using Codim0    = typename GridView::template Codim<0>;
    using PointType = typename Codim0::Geometry::GlobalCoordinate;
    using VectorOfPointsType = std::vector<PointType> ;
    
    using DataType = std::vector<Dune::FieldVector<double,1> >;
    using VectorOfDataType = std::vector<std::pair<std::hash<DataType>, DataType>>>;
    
    using Distance = nanoflann::metric_L2_Simple;
    
    using Self = KDTreeMeshAdaptor<VectorOfPointsType, VectorOfDataType, ValueType, dimensionworld, Distance, SizeType>;
    using metric_t = typename Distance::template traits<ValueType, Self>::distance_t;
    using index_t = KDTreeSingleIndexAdaptor<metric_t, Self, dimensionworld, SizeType>;

  public:
    unique_ptr<index_t> index; //! The kd-tree index for the user to call its methods as usual with any other FLANN index.
    FeSpace const& feSpace;

    VectorOfPointsType_ m_points;
    VectorOfDataType_ m_data;
    
    std::map<std::hash<DataType>, DataType> shapeValues;

  public:
    /// Constructor: takes a const ref to the vector of vectors object with the data points
    KDTreeMeshAdaptor(FeSpace const& feSpace_,
		      const int leaf_max_size = 10)
      : feSpace(feSpace_)
    {
      init();
      
      index = make_unique<index_t>(dimensionworld, *this /* adaptor */, 
                                        nanoflann::KDTreeSingleIndexAdaptorParams(leaf_max_size, dimensionworld));
      index->buildIndex();
    }

    /// Destructor.
    ~KDTreeMeshAdaptor() {}
    
    void init()
    {
      auto const& localView = feSpace.localView();
      auto const& indexSet = feSpace.localIndexSet();
      
      for (auto const& element : elements(feSpace.gridView())) {
        localView.bind(element);
        localIndexSet.bind(localView);
        
        m_points.push_back(element.geometry().center());
        
        auto const& localBasis = localView.tree().finiteElement().localBasis();
        std::vector<SizeType> indices(localBasis.size());
        for (size_t j = 0; j < localBasis.size(); ++j)
          indices[i] = localIndexSet.index(j);
        
        m_data.push_back(indices);
      }
    }
    
    void reinit(const int leaf_max_size = 10)
    {
      m_points.clear();
      m_data.clear();
      init();
      index.reset(new index_t(dimensionworld, *this /* adaptor */, 
                              nanoflann::KDTreeSingleIndexAdaptorParams(leaf_max_size, dimensionworld));
      index->buildIndex();
    }

    /** Query for the \a num_closest closest points to a given point (entered as query_point[0:dim-1]).
      *  Note that this is a short-cut method for index->findNeighbors().
      *  The user can also call index->... methods as desired.
      * \note nChecks_IGNORED is ignored but kept for compatibility with the original FLANN interface.
      */
    inline void query(const ValueType *query_point, 
		      const size_t num_closest, 
		      SizeType *out_indices, 
		      ValueType *out_distances_sq, 
		      const int nChecks_IGNORED = 10) const
    {
      nanoflann::KNNResultSet<ValueType, SizeType> resultSet(num_closest);
      resultSet.init(out_indices, out_distances_sq);
      index->findNeighbors(resultSet, query_point, nanoflann::SearchParams());
    }

    /** @name Interface expected by KDTreeSingleIndexAdaptor
      * @{ */

    Self const& derived() const 
    {
      return *this;
    }
    Self& derived() 
    {
      return *this;
    }

    // Must return the number of data points
    inline size_t kdtree_get_point_count() const 
    {
      return m_points.size();
    }

    // Returns the distance between the vector "p1[0:size-1]" and the data point with index "idx_p2" stored in the class:
    inline value_type kdtree_distance(const double *p1, 
				      const SizeType idx_p2, 
				      size_t size) const
    {
      double s = 0.0;
      for (size_t i = 0; i < size; i++) {
	const double d = p1[i] - m_points[idx_p2][i];
	s += d*d;
      }
      return s;
    }

    // Returns the dim'th component of the idx'th point in the class:
    inline value_type kdtree_get_pt(const SizeType idx, size_t dim) const 
    {
      return m_points[idx][dim];
    }

    // Optional bounding-box computation: return false to default to a standard bbox computation loop.
    //   Return true if the BBOX was already computed by the class and returned in "bb" so it can be avoided to redo it again.
    //   Look at bb.size() to find out the expected dimensionality (e.g. 2 or 3 for point clouds)
    template <class BBOX>
    bool kdtree_get_bbox(BBOX &bb) const 
    {
      return false;
    }

    /** @} */
    
    
    /**
    * strategy and nrOfPoints are ignored
    *  == evalAtPoint_simple
    **/
    template <class DOFVector>
    typename DOFVector::value_type 
    evalAtPoint(DOFVector const& vec, PointType const& x, int strategy = 0, int nrOfPoints = -1)
    {
      using T = typename DOFVector::value_type;
      T value = 0;
      
      std::vector<SizeType> indices;
      if (getNearestElement(x, indices)) {
        // get lambda of point...
        
        return localEval(vec, indices, lambda);
      }
      return value;
    }
    
  private:    
    /// evaluate DOFVector at barycentric coordinates \p lambda in element that
    /// is bind to instance in \ref bind().
    template <class FE, class ValueType, class Indices, class LocalCoord>
    ValueType localEval(DOFVector<FE, ValueType> const& vec, 
                        Indices const& indices, LocalCoord const& lambda)
    {
      auto const& localBasis = localView.tree().finiteElement().localBasis();
      
      // find a way to evaluate basisFcts at lambda!...
      auto const& shape = shapeValues[indices.first];
      localBasis.evaluateFunction(lambda, shapeValues);
      
      assert( shape.size() == indices.size() );
      
      ValueType data = 0;
      for (size_t j = 0; j < shape.size(); ++j)
        data += vec[indices.second[j]] * shape[j];
      
      return data;
    }
    
    template <class Indices>
    bool getNearestElement(PointType& x, Indices& indices)
    {
      const size_t nnp = 1;
      std::vector<DegreeOfFreedom> ret_indexes(nnp);
      std::vector<double> out_dists_sqr(nnp);
      nanoflann::KNNResultSet<double, DegreeOfFreedom> resultSet(nnp);
      
      resultSet.init(&ret_indexes[0], &out_dists_sqr[0]);
      index->findNeighbors(resultSet, x.begin(), nanoflann::SearchParams());
      
      indices = m_data[ret_indexes[0]];      
      return true;
    }

  }; // end of KDTreeMeshAdaptor

  

  template <class FeSpace>
  auto provideKDTree(FeSpace const& gridView)
  {
    using GridView = typename FeSpace::GridView;
    static constexpr int dow = GridView::dimensionworld;
    
    using KD_Tree = KDTreeMeshAdaptor<FeSpace, VectorOfPointsType, VectorOfDataType, double, dow>;

    return make_unique<KD_Tree_M>(dow, gridView);
  }
  

} }
