// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#include <iostream>

#include <dune/amdis/AMDiS.hpp>
#include <dune/amdis/AdaptInstationary.hpp>
#include <dune/amdis/ProblemInstat.hpp>
#include <dune/amdis/ProblemStat.hpp>
#include <dune/amdis/Terms.hpp>
#include <dune/amdis/common/Literals.hpp>

using namespace AMDiS;


class HeatParam
{
  template <class... Bs>
  using FeSpaceTuple = std::tuple<Bs...>;
  
  template <class Mesh, int deg>
  using Lagrange = Dune::Functions::PQkNodalBasis<typename Mesh::LeafGridView, deg>;
  
public:
  static constexpr int dim = AMDIS_DIM;
  static constexpr int dimworld = AMDIS_DOW;
  static constexpr int nComponents = 1;
  
  // default values
  using Mesh = Dune::UGGrid<dim>;
  using FeSpaces = FeSpaceTuple<Lagrange<Mesh, 2>>;
};


// 1 component with polynomial degree 1
// using HeatParam   = ProblemStatTraits<AMDIS_DIM, AMDIS_DOW, 1>;
using HeatProblem = ProblemStat<HeatParam>;
using HeatProblemInstat = ProblemInstat<HeatParam>;

int main(int argc, char** argv)
{
  AMDiS::init(argc, argv);
  
  HeatProblem prob("heat");
  prob.initialize(INIT_ALL);
  
  HeatProblemInstat probInstat("heat", prob);
  probInstat.initialize(INIT_UH_OLD);
  
  AdaptInfo adaptInfo("adapt");
  
  using Op = HeatProblem::OperatorType;
  Op opTimeLhs, opL, opTimeRhs, opForce; 
  
  opTimeLhs.addZOT( constant(1.0) );
  prob.addMatrixOperator(opTimeLhs, 0, 0, probInstat.getInvTau());
  
  opL.addSOT( constant(1.0) );
  prob.addMatrixOperator(opL, 0, 0);
  
  opTimeRhs.addZOT( valueOf(prob.getSolution(0_c)) );
  prob.addVectorOperator(opTimeRhs, 0, probInstat.getInvTau());
  
  opForce.addZOT( eval([](auto const& x) { return -1.0; }) );
  prob.addVectorOperator(opForce, 0);
  
  
  // set boundary condition
  auto predicate = [](auto const& p){ return p[0] < 1.e-8 || p[1] < 1.e-8; };
  auto dbcValues = [](auto const& p){ return 0.0; };
  prob.addDirichletBC(predicate, 0, 0, dbcValues);
      
  *prob.getSolution() = 0.0;
  
  AdaptInstationary adapt("adapt", prob, adaptInfo, probInstat, adaptInfo);
  adapt.adapt();
  
  AMDiS::finalize();
  return 0;
}